@@ -12,6 +12,7 @@
 
 import netaddr
 import testscenarios
+import uuid
 
 from ovsdbapp.backend.ovs_idl import idlutils
 from ovsdbapp import constants as const
@@ -36,7 +37,8 @@ class TestLogicalSwitchOps(OvnNorthboundTest):
         self.table = self.api.tables['Logical_Switch']
 
     def _ls_add(self, *args, **kwargs):
-        fix = self.useFixture(fixtures.LogicalSwitchFixture(*args, **kwargs))
+        fix = self.useFixture(fixtures.LogicalSwitchFixture(self.api, *args,
+                                                            **kwargs))
         self.assertIn(fix.obj.uuid, self.table.rows)
         return fix.obj
 
@@ -97,8 +99,7 @@ class TestLogicalSwitchOps(OvnNorthboundTest):
         self.api.ls_del(name, if_exists=True).execute(check_error=True)
 
     def test_ls_list(self):
-        with self.api.transaction(check_error=True):
-            switches = {self._ls_add() for _ in range(3)}
+        switches = {self._ls_add() for _ in range(3)}
         switch_set = set(self.api.ls_list().execute(check_error=True))
         self.assertTrue(switches.issubset(switch_set))
 
@@ -106,8 +107,10 @@ class TestLogicalSwitchOps(OvnNorthboundTest):
 class TestAclOps(OvnNorthboundTest):
     def setUp(self):
         super(TestAclOps, self).setUp()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
-        self.port_group = self.useFixture(fixtures.PortGroupFixture()).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api)).obj
+        self.port_group = self.useFixture(
+            fixtures.PortGroupFixture(self.api)).obj
 
     def _acl_add(self, entity, *args, **kwargs):
         self.assertIn(entity, ['lswitch', 'port_group'])
@@ -198,6 +201,24 @@ class TestAclOps(OvnNorthboundTest):
         self._acl_add('port_group', 'from-lport', 0,
                       'output == "fake_port" && ip', 'drop')
 
+    def test_pg_acl_add_logging(self):
+        r1 = self._acl_add('port_group', 'from-lport', 0,
+                           'output == "fake_port" && ip', 'drop',
+                           log=True, severity="warning", name="test1",
+                           meter="meter1", extkey1="extval1")
+        self.assertTrue(r1.log)
+        self.assertEqual(["meter1"], r1.meter)
+        self.assertEqual(["warning"], r1.severity)
+        self.assertEqual(["test1"], r1.name)
+        self.assertEqual({"extkey1": "extval1"}, r1.external_ids)
+
+    def test_pg_acl_add_logging_weird_severity(self):
+        r1 = self._acl_add('port_group', 'from-lport', 0,
+                           'output == "fake_port" && ip', 'drop',
+                           severity="weird")
+        self.assertNotEqual(["weird"], r1.severity)
+        self.assertEqual([], r1.severity)
+
     def test_pg_acl_del_all(self):
         r1 = self._acl_add('port_group', 'from-lport', 0,
                            'output == "fake_port"', 'drop')
@@ -219,7 +240,8 @@ class TestAclOps(OvnNorthboundTest):
 class TestQoSOps(OvnNorthboundTest):
     def setUp(self):
         super(TestQoSOps, self).setUp()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api)).obj
 
     def _qos_add(self, *args, **kwargs):
         cmd = self.api.qos_add(self.switch.uuid, *args, **kwargs)
@@ -231,7 +253,7 @@ class TestQoSOps(OvnNorthboundTest):
         self.assertEqual(cmd.rate, row.bandwidth.get('rate', None))
         self.assertEqual(cmd.burst, row.bandwidth.get('burst', None))
         self.assertEqual(cmd.dscp, row.action.get('dscp', None))
-        return row
+        return idlutils.frozen_row(row)
 
     def test_qos_add_dscp(self):
         self._qos_add('from-lport', 0, 'output == "fake_port" && ip', dscp=33)
@@ -266,10 +288,81 @@ class TestQoSOps(OvnNorthboundTest):
 
     def test_qos_add_may_exist(self):
         args = ('from-lport', 0, 'output == "fake_port" && ip', 1000)
-        row = self._qos_add(*args)
-        row2 = self._qos_add(*args, may_exist=True)
+        row = self._qos_add(*args, external_ids={'port_id': '1'})
+        row2 = self._qos_add(*args, external_ids={'port_id': '1'},
+                             may_exist=True)
         self.assertEqual(row, row2)
 
+    def test_qos_add_may_exist_update_using_external_ids_match(self):
+        args = ('from-lport', 0, 'output == "fake_port" && ip')
+        kwargs = {'rate': 1000, 'burst': 800, 'dscp': 16,
+                  'external_ids': {'port_id': '1'}}
+        row = self._qos_add(*args, **kwargs)
+
+        # Update QoS parameters: rate, burst and DSCP.
+        kwargs = {'rate': 1200, 'burst': 900, 'dscp': 24}
+        row2 = self._qos_add(*args, external_ids_match={'port_id': '1'},
+                             may_exist=True, **kwargs)
+        self.assertEqual(row.uuid, row2.uuid)
+        self.assertEqual(row2.bandwidth, {'rate': 1200, 'burst': 900})
+        self.assertEqual(row2.action, {'dscp': 24})
+
+        # Remove QoS parameters.
+        kwargs = {'rate': 1500, 'burst': 1100}
+        row3 = self._qos_add(*args, external_ids_match={'port_id': '1'},
+                             may_exist=True, **kwargs)
+        self.assertEqual(row.uuid, row3.uuid)
+        self.assertEqual(row3.bandwidth, {'rate': 1500, 'burst': 1100})
+        self.assertEqual(row3.action, {})
+
+        kwargs = {'rate': 2000}
+        row4 = self._qos_add(*args, external_ids_match={'port_id': '1'},
+                             may_exist=True, **kwargs)
+        self.assertEqual(row.uuid, row4.uuid)
+        self.assertEqual(row4.bandwidth, {'rate': 2000})
+        self.assertEqual(row4.action, {})
+
+        kwargs = {'dscp': 16}
+        row5 = self._qos_add(*args, external_ids_match={'port_id': '1'},
+                             may_exist=True, **kwargs)
+        self.assertEqual(row.uuid, row5.uuid)
+        self.assertEqual(row5.bandwidth, {})
+        self.assertEqual(row5.action, {'dscp': 16})
+
+    def test_qos_add_may_exist_using_external_ids_match(self):
+        _uuid = str(uuid.uuid4())
+        _uuid2 = str(uuid.uuid4())
+        for key in ('neutron:port_id', 'neutron:fip_id'):
+            args = ('from-lport', 0, 'output == "fake_port" && ip')
+            kwargs = {'rate': 1000, 'burst': 800, 'dscp': 16}
+            self._qos_add(*args, external_ids={key: _uuid}, **kwargs)
+
+            # "args" in this second call are different, "QoSAddCommand" will
+            # use the "external_ids_match" reference passed instead to match
+            # the QoS rules.
+            args = ('from-lport', 1, 'output == "fake_port" && ip')
+            self._qos_add(*args, external_ids_match={key: _uuid},
+                          may_exist=True, **kwargs)
+            qos_rules = self.api.qos_list(self.switch.uuid).execute(
+                check_error=True)
+            self.assertEqual(1, len(qos_rules))
+
+            # This call will update the "external_ids" to "_uuid2". Before
+            # changing it, "_uuid" will be used to find the register.
+            self._qos_add(*args, external_ids_match={key: _uuid},
+                          external_ids={key: _uuid2}, may_exist=True, **kwargs)
+            qos_rules = self.api.qos_list(self.switch.uuid).execute(
+                check_error=True)
+            self.assertEqual(1, len(qos_rules))
+
+            # The deletion call uses "_uuid2" because it was changed in the
+            # previous call.
+            self.api.qos_del_ext_ids(self.switch.name,
+                                     {key: _uuid2}).execute(check_error=True)
+            qos_rules = self.api.qos_list(self.switch.uuid).execute(
+                check_error=True)
+            self.assertEqual(0, len(qos_rules))
+
     def test_qos_add_extids(self):
         external_ids = {'mykey': 'myvalue', 'yourkey': 'yourvalue'}
         qos = self._qos_add('from-lport', 0, 'output == "fake_port" && ip',
@@ -287,8 +380,9 @@ class TestQoSOps(OvnNorthboundTest):
         r2 = self._qos_add('to-lport', 0, 'output == "fake_port"', 1000)
         self.api.qos_del(self.switch.uuid, 'from-lport').execute(
             check_error=True)
-        self.assertNotIn(r1, self.switch.qos_rules)
-        self.assertIn(r2, self.switch.qos_rules)
+        qos_rules = [idlutils.frozen_row(row) for row in self.switch.qos_rules]
+        self.assertNotIn(r1, qos_rules)
+        self.assertIn(r2, qos_rules)
 
     def test_qos_del_direction_priority_match(self):
         r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
@@ -296,8 +390,9 @@ class TestQoSOps(OvnNorthboundTest):
         cmd = self.api.qos_del(self.switch.uuid,
                                'from-lport', 0, 'output == "fake_port"')
         cmd.execute(check_error=True)
-        self.assertNotIn(r1, self.switch.qos_rules)
-        self.assertIn(r2, self.switch.qos_rules)
+        qos_rules = [idlutils.frozen_row(row) for row in self.switch.qos_rules]
+        self.assertNotIn(r1, qos_rules)
+        self.assertIn(r2, qos_rules)
 
     def test_qos_del_priority_without_match(self):
         self.assertRaises(TypeError, self.api.qos_del, self.switch.uuid,
@@ -307,21 +402,98 @@ class TestQoSOps(OvnNorthboundTest):
         self.assertRaises(TypeError, self.api.qos_del, self.switch.uuid,
                           priority=0)
 
+    def test_qos_del_ls_not_present_if_exists_true(self):
+        self.api.qos_del('some_other_ls').execute(check_error=True)
+
+    def test_qos_del_ls_not_present_if_exists_false(self):
+        cmd = self.api.qos_del('some_other_ls', if_exists=False)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
     def test_qos_list(self):
         r1 = self._qos_add('from-lport', 0, 'output == "fake_port"', 1000)
         r2 = self._qos_add('from-lport', 1, 'output == "fake_port2"', 1000)
         qos_rules = self.api.qos_list(self.switch.uuid).execute(
             check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
         self.assertIn(r1, qos_rules)
         self.assertIn(r2, qos_rules)
 
+    def _create_fip_qoses(self):
+        ext_ids_1 = {'key1': 'value1', 'key2': 'value2'}
+        self.qos_1 = self._qos_add('from-lport', 0, 'output == "fake_port1"',
+                                   dscp=11, external_ids=ext_ids_1)
+        ext_ids_2 = {'key3': 'value3', 'key4': 'value4'}
+        self.qos_2 = self._qos_add('from-lport', 1, 'output == "fake_port2"',
+                                   dscp=11, external_ids=ext_ids_2)
+        self.qos_3 = self._qos_add('from-lport', 2, 'output == "fake_port3"',
+                                   dscp=10)
+
+    def test_qos_delete_external_ids(self):
+        self._create_fip_qoses()
+        self.api.qos_del_ext_ids(self.switch.name,
+                                 {'key1': 'value1'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
+        self.assertCountEqual([self.qos_2, self.qos_3], qos_rules)
+
+        self.api.qos_del_ext_ids(
+            self.switch.name,
+            {'key3': 'value3', 'key4': 'value4'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
+        self.assertCountEqual([self.qos_3], qos_rules)
+
+    def test_qos_delete_external_ids_wrong_keys_or_values(self):
+        self._create_fip_qoses()
+        self.api.qos_del_ext_ids(self.switch.name,
+                                 {'key_z': 'value1'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
+        self.api.qos_del_ext_ids(self.switch.name,
+                                 {'key1': 'value_z'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
+        self.api.qos_del_ext_ids(
+            self.switch.name,
+            {'key3': 'value3', 'key4': 'value_z'}).execute(check_error=True)
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
+    def test_qos_delete_external_ids_empty_dict(self):
+        self.assertRaises(TypeError, self.api.qos_del_ext_ids,
+                          self.switch.name, {})
+
+    def test_qos_delete_external_ids_if_exists(self):
+        self._create_fip_qoses()
+        cmd = self.api.qos_del_ext_ids('wrong_ls_name',
+                                       {'key1': 'value1'}, if_exists=False)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+        self.api.qos_del_ext_ids('wrong_ls_name',
+                                 {'key1': 'value1'}).execute(check_error=True)
+        # No qos rule has been deleted from the correct logical switch.
+        qos_rules = self.api.qos_list(
+            self.switch.uuid).execute(check_error=True)
+        qos_rules = [idlutils.frozen_row(row) for row in qos_rules]
+        self.assertCountEqual([self.qos_1, self.qos_2, self.qos_3], qos_rules)
+
 
 class TestLspOps(OvnNorthboundTest):
     def setUp(self):
         super(TestLspOps, self).setUp()
         name = utils.get_rand_device_name()
         self.switch = self.useFixture(
-            fixtures.LogicalSwitchFixture(name)).obj
+            fixtures.LogicalSwitchFixture(self.api, name)).obj
 
     def _lsp_add(self, switch, name, *args, **kwargs):
         name = utils.get_rand_device_name() if name is None else name
@@ -356,7 +528,7 @@ class TestLspOps(OvnNorthboundTest):
         self.assertEqual(lsp1, lsp2)
 
     def test_lsp_add_may_exist_wrong_switch(self):
-        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        sw = self.useFixture(fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self._lsp_add(self.switch, None)
         self.assertRaises(RuntimeError, self._lsp_add, sw, lsp.name,
                           may_exist=True)
@@ -435,7 +607,7 @@ class TestLspOps(OvnNorthboundTest):
 
     def test_lsp_del_wrong_switch(self):
         lsp = self._lsp_add(self.switch, None)
-        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        sw = self.useFixture(fixtures.LogicalSwitchFixture(self.api)).obj
         cmd = self.api.lsp_del(lsp.uuid, sw.uuid)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
 
@@ -460,7 +632,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_list_no_switch(self):
         ports = {self._lsp_add(self.switch, None) for _ in range(3)}
         other_switch = self.useFixture(fixtures.LogicalSwitchFixture(
-            name=utils.get_rand_device_name())).obj
+            self.api, name=utils.get_rand_device_name())).obj
         other_port = self._lsp_add(other_switch, None)
         all_ports = set(self.api.lsp_list().execute(check_error=True))
         self.assertTrue((ports.union(set([other_port]))).issubset(all_ports))
@@ -547,7 +719,7 @@ class TestLspOps(OvnNorthboundTest):
     def test_lsp_set_get_dhcpv4_options(self):
         lsp = self._lsp_add(self.switch, None)
         dhcpopt = self.useFixture(
-            fixtures.DhcpOptionsFixture('192.0.2.1/24')).obj
+            fixtures.DhcpOptionsFixture(self.api, '192.0.2.1/24')).obj
         self.api.lsp_set_dhcpv4_options(
             lsp.name, dhcpopt.uuid).execute(check_error=True)
         options = self.api.lsp_get_dhcpv4_options(
@@ -558,7 +730,7 @@ class TestLspOps(OvnNorthboundTest):
 class TestDhcpOptionsOps(OvnNorthboundTest):
     def _dhcpopt_add(self, cidr, *args, **kwargs):
         dhcpopt = self.useFixture(fixtures.DhcpOptionsFixture(
-            cidr, *args, **kwargs)).obj
+            self.api, cidr, *args, **kwargs)).obj
         self.assertEqual(cidr, dhcpopt.cidr)
         return dhcpopt
 
@@ -605,7 +777,7 @@ class TestDhcpOptionsOps(OvnNorthboundTest):
 class TestLogicalRouterOps(OvnNorthboundTest):
     def _lr_add(self, *args, **kwargs):
         lr = self.useFixture(
-            fixtures.LogicalRouterFixture(*args, **kwargs)).obj
+            fixtures.LogicalRouterFixture(self.api, *args, **kwargs)).obj
         self.assertIn(lr.uuid, self.api.tables['Logical_Router'].rows)
         return lr
 
@@ -733,7 +905,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 
     def test_lr_nat_add_port(self):
         sw = self.useFixture(
-            fixtures.LogicalSwitchFixture()).obj
+            fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self.api.lsp_add(sw.uuid, utils.get_rand_device_name()).execute(
             check_error=True)
         lport, mac = (lsp.name, 'de:ad:be:ef:4d:ad')
@@ -761,7 +933,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 
     def test_lr_nat_add_may_exist(self):
         sw = self.useFixture(
-            fixtures.LogicalSwitchFixture()).obj
+            fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self.api.lsp_add(sw.uuid, utils.get_rand_device_name()).execute(
             check_error=True)
         args = (const.NAT_BOTH, '10.17.4.1', '192.0.2.1')
@@ -776,7 +948,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 
     def test_lr_nat_add_may_exist_remove_port(self):
         sw = self.useFixture(
-            fixtures.LogicalSwitchFixture()).obj
+            fixtures.LogicalSwitchFixture(self.api)).obj
         lsp = self.api.lsp_add(sw.uuid, utils.get_rand_device_name()).execute(
             check_error=True)
         args = (const.NAT_BOTH, '10.17.4.1', '192.0.2.1')
@@ -878,7 +1050,7 @@ class TestLogicalRouterOps(OvnNorthboundTest):
 class TestLogicalRouterPortOps(OvnNorthboundTest):
     def setUp(self):
         super(TestLogicalRouterPortOps, self).setUp()
-        self.lr = self.useFixture(fixtures.LogicalRouterFixture()).obj
+        self.lr = self.useFixture(fixtures.LogicalRouterFixture(self.api)).obj
 
     def _lrp_add(self, port, mac='de:ad:be:ef:4d:ad',
                  networks=None, *args, **kwargs):
@@ -932,7 +1104,7 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
     def test_lrp_add_may_exist_different_router(self):
         name = utils.get_rand_device_name()
         args = (name, 'de:ad:be:ef:4d:ad', ['192.0.2.0/24'])
-        lr2 = self.useFixture(fixtures.LogicalRouterFixture()).obj
+        lr2 = self.useFixture(fixtures.LogicalRouterFixture(self.api)).obj
         self._lrp_add(*args)
         cmd = self.api.lrp_add(lr2.uuid, *args, may_exist=True)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
@@ -998,7 +1170,7 @@ class TestLogicalRouterPortOps(OvnNorthboundTest):
 
     def test_lrp_del_wrong_router(self):
         lrp = self._lrp_add(None)
-        sw = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        sw = self.useFixture(fixtures.LogicalSwitchFixture(self.api)).obj
         cmd = self.api.lrp_del(lrp.uuid, sw.uuid)
         self.assertRaises(RuntimeError, cmd.execute, check_error=True)
 
@@ -1046,7 +1218,7 @@ class TestLoadBalancerOps(OvnNorthboundTest):
     def _lb_add(self, lb, vip, ips, protocol=const.PROTO_TCP, may_exist=False,
                 **columns):
         lbal = self.useFixture(fixtures.LoadBalancerFixture(
-            lb, vip, ips, protocol, may_exist, **columns)).obj
+            self.api, lb, vip, ips, protocol, may_exist, **columns)).obj
         self.assertEqual(lb, lbal.name)
         norm_vip = ovsdb_utils.normalize_ip_port(vip)
         self.assertIn(norm_vip, lbal.vips)
@@ -1126,6 +1298,32 @@ class TestLoadBalancerOps(OvnNorthboundTest):
         self.api.lb_del(utils.get_rand_device_name(), if_exists=True).execute(
             check_error=True)
 
+    def test_lb_del_vip_if_exists(self):
+        name = utils.get_rand_device_name()
+        lb = self._lb_add(name, '192.0.2.1', ['10.0.0.1'])
+        lb2 = self._lb_add(name, '192.0.2.2', ['10.1.0.1'])
+        self.assertEqual(lb, lb2)
+        # Remove vip that does not exist.
+        self.api.lb_del(lb.name, '9.9.9.9', if_exists=True
+                        ).execute(check_error=True)
+        self.assertIn('192.0.2.1', lb.vips)
+        self.assertIn('192.0.2.2', lb.vips)
+        # Remove one vip.
+        self.api.lb_del(lb.name, '192.0.2.1', if_exists=True
+                        ).execute(check_error=True)
+        self.assertNotIn('192.0.2.1', lb.vips)
+        # Attempt to remove same vip a second time with if_exists=True.
+        self.api.lb_del(lb.name, '192.0.2.1', if_exists=True).execute(
+            check_error=True)
+        # Attempt to remove same vip a third time with if_exists=False.
+        cmd = self.api.lb_del(lb.name, '192.0.2.1', if_exists=False)
+        self.assertRaises(idlutils.RowNotFound, cmd.execute, check_error=True)
+        # Remove last vip and make sure that lb is also removed.
+        self.assertIn('192.0.2.2', lb.vips)
+        self.api.lb_del(lb.name, '192.0.2.2').execute(check_error=True)
+        cmd = self.api.lb_del(lb.name)
+        self.assertRaises(idlutils.RowNotFound, cmd.execute, check_error=True)
+
     def test_lb_list(self):
         lbs = {self._lb_add(utils.get_rand_device_name(), '192.0.2.1',
                             ['10.0.0.1', '10.0.0.2']) for _ in range(3)}
@@ -1151,13 +1349,13 @@ class TestObLbOps(testscenarios.TestWithScenarios, OvnNorthboundTest):
         # They must be in this order because the load balancer
         # can't be deleted when there is a reference in the router
         self.lb = self.useFixture(fixtures.LoadBalancerFixture(
-            utils.get_rand_device_name(), '192.0.2.1',
+            self.api, utils.get_rand_device_name(), '192.0.2.1',
             ['10.0.0.1', '10.0.0.2'])).obj
         self.lb2 = self.useFixture(fixtures.LoadBalancerFixture(
-            utils.get_rand_device_name(), '192.0.2.2',
+            self.api, utils.get_rand_device_name(), '192.0.2.2',
             ['10.1.0.1', '10.1.0.2'])).obj
         self.lr = self.useFixture(self.fixture(
-            utils.get_rand_device_name())).obj
+            self.api, utils.get_rand_device_name())).obj
 
     def test_ob_lb_add(self):
         self.add_fn(self.lr.name, self.lb.name).execute(
@@ -1214,7 +1412,8 @@ class TestCommonDbOps(OvnNorthboundTest):
     def setUp(self):
         super(TestCommonDbOps, self).setUp()
         name = utils.get_rand_device_name()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture(name)).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api, name)).obj
         self.lsps = [
             self.api.lsp_add(
                 self.switch.uuid,
@@ -1272,7 +1471,8 @@ class TestCommonDbOps(OvnNorthboundTest):
 
 class TestDnsOps(OvnNorthboundTest):
     def _dns_add(self, *args, **kwargs):
-        dns = self.useFixture(fixtures.DnsFixture(*args, **kwargs)).obj
+        dns = self.useFixture(
+            fixtures.DnsFixture(self.api, *args, **kwargs)).obj
         return dns
 
     def test_dns_get(self):
@@ -1340,11 +1540,13 @@ class TestDnsOps(OvnNorthboundTest):
 
 class TestLsDnsOps(OvnNorthboundTest):
     def _dns_add(self, *args, **kwargs):
-        dns = self.useFixture(fixtures.DnsFixture(*args, **kwargs)).obj
+        dns = self.useFixture(
+            fixtures.DnsFixture(self.api, *args, **kwargs)).obj
         return dns
 
     def _ls_add(self, *args, **kwargs):
-        fix = self.useFixture(fixtures.LogicalSwitchFixture(*args, **kwargs))
+        fix = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api, *args, **kwargs))
         return fix.obj
 
     def test_ls_dns_set_clear_records(self):
@@ -1353,7 +1555,7 @@ class TestLsDnsOps(OvnNorthboundTest):
 
         ls1 = self._ls_add('ls1')
         self.api.ls_set_dns_records(ls1.uuid, [dns1.uuid, dns2.uuid]).execute()
-        self.assertItemsEqual([dns1.uuid, dns2.uuid],
+        self.assertCountEqual([dns1.uuid, dns2.uuid],
                               [dns.uuid for dns in ls1.dns_records])
 
         self.api.ls_clear_dns_records(ls1.uuid).execute()
@@ -1366,15 +1568,15 @@ class TestLsDnsOps(OvnNorthboundTest):
         ls1 = self._ls_add('ls1')
 
         self.api.ls_add_dns_record(ls1.uuid, dns1.uuid).execute()
-        self.assertItemsEqual([dns1.uuid],
+        self.assertCountEqual([dns1.uuid],
                               [dns.uuid for dns in ls1.dns_records])
 
         self.api.ls_add_dns_record(ls1.uuid, dns2.uuid).execute()
-        self.assertItemsEqual([dns1.uuid, dns2.uuid],
+        self.assertCountEqual([dns1.uuid, dns2.uuid],
                               [dns.uuid for dns in ls1.dns_records])
 
         self.api.ls_remove_dns_record(ls1.uuid, dns2.uuid).execute()
-        self.assertItemsEqual([dns1.uuid],
+        self.assertCountEqual([dns1.uuid],
                               [dns.uuid for dns in ls1.dns_records])
         self.api.ls_remove_dns_record(ls1.uuid, dns1.uuid).execute()
         self.assertEqual([], ls1.dns_records)
@@ -1384,7 +1586,8 @@ class TestPortGroup(OvnNorthboundTest):
 
     def setUp(self):
         super(TestPortGroup, self).setUp()
-        self.switch = self.useFixture(fixtures.LogicalSwitchFixture()).obj
+        self.switch = self.useFixture(
+            fixtures.LogicalSwitchFixture(self.api)).obj
         self.pg_name = 'testpg-%s' % ovsdb_utils.generate_uuid()
 
     def test_port_group(self):
@@ -1587,3 +1790,89 @@ class TestReferencedObjects(OvnNorthboundTest):
             txn.add(self.api.db_add('Logical_Switch_Port', prt,
                                     "addresses", self.lsp_test_addresses[0]))
         self._check_values()
+
+
+class TestMeterOps(OvnNorthboundTest):
+
+    def setUp(self):
+        super(TestMeterOps, self).setUp()
+        self.table = self.api.tables['Meter']
+
+    def _meter_add(self, name=None, unit=None, rate=1, fair=False,
+                   burst_size=0, action=None, may_exist=False, **columns):
+        if name is None:
+            name = utils.get_rand_name()
+        if unit is None:
+            unit = 'pktps'
+        # NOTE(flaviof): action drop is the only one option in the schema when
+        #                this test was implemented. That is expected to be
+        #                properly handled by api.
+        exp_action = 'drop' if action is None else action
+        meter = self.useFixture(fixtures.MeterFixture(
+            self.api, name, unit, rate, fair, burst_size, action, may_exist,
+            **columns)).obj
+        meter_band = meter.bands[0]
+        self.assertEqual(name, meter.name)
+        self.assertEqual(unit, meter.unit)
+        self.assertEqual([fair], meter.fair)
+        self.assertEqual(exp_action, meter_band.action)
+        self.assertEqual(rate, meter_band.rate)
+        self.assertEqual(burst_size, meter_band.burst_size)
+        return meter
+
+    def _meter_del(self, row, col):
+        name404 = utils.get_rand_name()
+        self.api.meter_del(name404, if_exists=True).execute(check_error=True)
+        self.assertIn(row.uuid, self.table.rows)
+        self.api.meter_del(row.uuid).execute(check_error=True)
+        self.assertNotIn(row.uuid, self.table.rows)
+        self.api.meter_del(col, if_exists=True).execute(check_error=True)
+        cmd = self.api.meter_del(col)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+    def _meter_get(self, col):
+        meter = self._meter_add()
+        val = getattr(meter, col)
+        found = self.api.meter_get(val).execute(check_error=True)
+        self.assertEqual(meter, found)
+
+    def test_meter_add_non_defaults(self):
+        self._meter_add(utils.get_rand_name(), "kbps", 321, True, 500, None)
+
+    def test_meter_add_ext_ids(self):
+        ext_ids = {
+            utils.get_rand_name(prefix="random_"): utils.get_random_string(10)
+            for _ in range(3)}
+        meter = self._meter_add(external_ids=ext_ids)
+        self.assertEqual(ext_ids, meter.external_ids)
+
+    def test_meter_add_may_exist(self):
+        cmd = self.api.meter_add("same_name", "kbps", may_exist=True)
+        m1 = cmd.execute(check_error=True)
+        m2 = cmd.execute(check_error=True)
+        self.assertEqual(m1, m2)
+
+    def test_meter_add_duplicate(self):
+        cmd = self.api.meter_add("same_name", "kbps", may_exist=False)
+        cmd.execute(check_error=True)
+        self.assertRaises(RuntimeError, cmd.execute, check_error=True)
+
+    def test_meter_del_by_uuid(self):
+        meter = self._meter_add()
+        self._meter_del(meter, meter.uuid)
+
+    def test_meter_del_by_name(self):
+        name = utils.get_rand_name()
+        meter = self._meter_add(name)
+        self._meter_del(meter, name)
+
+    def test_meter_list(self):
+        meters = {self._meter_add() for _ in range(3)}
+        meter_set = set(self.api.meter_list().execute(check_error=True))
+        self.assertTrue(meters.issubset(meter_set))
+
+    def test_meter_get_uuid(self):
+        self._meter_get('uuid')
+
+    def test_meter_get_name(self):
+        self._meter_get('name')
